/* id3attr.cpp - main file
 *
 * id3attr (c) 2000-2001 pinc Software, Axel Drfler (axeld@pinc-software.de)
 *
 * based on id3Attr (c) 1999 Ari Pernick (pernick@uiuc.edu) which is:
 * Based very partially on ID3ren (by Robert Alto (badcrc@tscnet.com)
 * and its BeOS port by Shayne White (shayne@curvedspace.org))
 * Which was based on (c) 1999 Pieter Panman (ppanman@dds.nl)
 * 
 * This program is free software you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

#include <Application.h>
#include <Path.h>
#include <File.h>
#include <NodeInfo.h>
#include <Node.h>
#include <Directory.h>
#include <InterfaceDefs.h>
#include <UTF8.h>
#include <kernel/fs_info.h>
#include <kernel/fs_index.h>
#include <kernel/fs_attr.h>

#include <MediaFile.h>
#include <MediaTrack.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "id3attr.h"
#include "genre.h"

bool recursive = false;		// enter directories recursively
bool force = false;			// overwrite existing attributes
bool frameRate = false;		// set framerate
bool leadingZero = false;	// precedes the length with a '0' instead of with a blank
bool upperArtist = false;
bool upperTitle = false;
bool upperAlbum = false;
bool useMediaKit = true;


void
upperString(char *t)
{
	bool word = false;

	for (; *t; t++) {
		if (isspace(*t))
			word = false;
		else if (!word)
			*t = toupper(*t), word = true;
	}
}


void
freeAudioAttributes(audio_attrs *audioAttrs)
{
	if (audioAttrs == NULL)
		return;

	free(audioAttrs->title);
	free(audioAttrs->artist);
	free(audioAttrs->album);
	free(audioAttrs->comment);
}


bool
isWriteAllowed(BNode *node,char *attr)
{
	attr_info attrInfo;
	
	return (force || node->GetAttrInfo(attr,&attrInfo) == B_ENTRY_NOT_FOUND);
}


void
writeAttributeString(BNode *node, char *attr, char *name)
{
	if (isWriteAllowed(node,attr))
		node->WriteAttr(attr, B_STRING_TYPE, 0, name, strlen(name) + 1);
}


void
writeAudioAttributes(BEntry *entry, audio_attrs *audioAttrs)
{
	BNode node(entry);
	
	if (upperArtist)
		upperString(audioAttrs->artist);
	writeAttributeString(&node, "Audio:Artist", audioAttrs->artist);
	if (upperTitle)
		upperString(audioAttrs->title);
	writeAttributeString(&node, "Audio:Title", audioAttrs->title);
	if (upperAlbum)
		upperString(audioAttrs->album);
	writeAttributeString(&node, "Audio:Album", audioAttrs->album);
	
	if (isWriteAllowed(&node, "Audio:Year"))
		node.WriteAttr("Audio:Year", B_INT32_TYPE, 0, (void *)&audioAttrs->year, sizeof(int32));
		
	writeAttributeString(&node, "Audio:Comment", audioAttrs->comment);
	writeAttributeString(&node, "Audio:Genre", audioAttrs->genre);

	if (audioAttrs->track != 0 && isWriteAllowed(&node, "Audio:Track"))
		node.WriteAttr("Audio:Track", B_INT32_TYPE, 0, (void *)&audioAttrs->track, sizeof(int32));

	char buffer[32];

	if (audioAttrs->bitrate) {
		sprintf(buffer, "%ld kbit", audioAttrs->bitrate);
		writeAttributeString(&node, "Audio:Bitrate", buffer);
	}
	if (audioAttrs->length) {
		int32 min = audioAttrs->length / 60000000;
		int32 sec = ((audioAttrs->length % 60000000)+500000) / 1000000;
		char *format = "%2ld:%02ld";

		if (leadingZero)
			format = "%02ld:%02ld";

		sprintf(buffer, format, min, sec);
		writeAttributeString(&node, "Audio:Length", buffer);
	}
	if (audioAttrs->framerate && frameRate) {
		sprintf(buffer, "%ld Hz", (int32)(audioAttrs->framerate + 0.5));
		writeAttributeString(&node, "Audio:Framerate", buffer);
	}
}


char *
convertToUTF8(char *t)
{
	char s[256];
	int32 sourceLen = strlen(t)+1,destLen = 255,state = 0;
	
	convert_to_utf8(B_ISO1_CONVERSION,t,&sourceLen,s,&destLen,&state);
	return(strdup(s));
}


void checkString(char *t)
{
	char *s = t;
	
	while(*s && isspace(*s))
		s++;
	
	if (*s) {
		s = t+strlen(t)-1;
		while(isspace(*s))
			*s-- = '\0';
	} else // string contains only spaces
		*t = '\0';
}


void
retrieveFromFile(char *name, BFile *file, audio_attrs *audioAttrs)
{
	id3struct id3tag = {"\0","\0","\0","\0","\0","\0",155};
	char buf[128] = "\0";

	// Read from file into buf
	file->Seek(-128, SEEK_END);
	file->Read((void *)buf, 128);

	// If first part of buf = "TAG" (it has a tag), split buf up into id3 tag parts
	if (!strncmp(buf, "TAG", 3)) {
		memcpy(id3tag.tag     , buf      , 3 );
		memcpy(id3tag.title   , buf + 3  , 30);
		memcpy(id3tag.artist  , buf + 33 , 30);
		memcpy(id3tag.album   , buf + 63 , 30);
		memcpy(id3tag.year    , buf + 93 , 4 );
		memcpy(id3tag.comment , buf + 97 , 30);
		id3tag.genre = buf[127];
	}
	checkString(id3tag.title);
	checkString(id3tag.artist);
	checkString(id3tag.album);
	checkString(id3tag.comment);

	audioAttrs->title = convertToUTF8(id3tag.title);
	audioAttrs->artist = convertToUTF8(id3tag.artist);
	audioAttrs->album = convertToUTF8(id3tag.album);
	audioAttrs->comment = convertToUTF8(id3tag.comment);
	audioAttrs->year = atoi(id3tag.year);
	if (id3tag.genre >= 0 && id3tag.genre < genre_count)
		audioAttrs->genre = genre_table[id3tag.genre];
	else
		audioAttrs->genre = "Other";

	if (id3tag.comment[28] == 0)
		audioAttrs->track = id3tag.comment[29];

	/* try to parse file name */
	
	if (!id3tag.title[0] || !id3tag.artist[0]) {
		char t[B_FILE_NAME_LENGTH];
		int i;
		
		strcpy(t,name);
		for(i = 0;t[i];i++)
			if (t[i] == '_')
				t[i] = ' ';
		for(i = 0;t[i] && t[i] != '-';i++);
		
		char *last;
		
		if (t[i] == '-') {
			char *s = t+i;

			*s = '\0';
			while (*--s == ' ' && s >= t) // remove trailing spaces
				*s = '\0';

			if (!id3tag.artist[0])
				audioAttrs->artist = strdup(t);

			last = t+i+1;
		} else
			last = t;

		while (*last == ' ')
			last++;
			
		if (!strncasecmp(last + (i = strlen(last) - 4), ".mp", 3))
			last[i] = '\0';

		if (!id3tag.title[0])
			audioAttrs->title = strdup(last);
	}

	/********************* Retrieve bit rate and length *********************/

	if (!useMediaKit)
		return;

	file->Seek(0, SEEK_SET);
	BMediaFile *mediaFile = new BMediaFile(file);
	
	if (mediaFile->InitCheck() == B_OK) {
		BMediaTrack *track;
		int32 numTracks,i;

		numTracks = mediaFile->CountTracks();

		for (i = 0; i < numTracks; i++) {
			if ((track = mediaFile->TrackAt(i)) == NULL)
				continue;

			media_format format;
			if (track->EncodedFormat(&format) == B_OK && format.type == B_MEDIA_ENCODED_AUDIO) {
				audioAttrs->bitrate = (int32)(format.u.encoded_audio.bit_rate / 1000);
				audioAttrs->length = track->Duration();
				audioAttrs->framerate = format.u.encoded_audio.output.frame_rate;
			}
			mediaFile->ReleaseTrack(track);
		}
	}
	delete mediaFile;
}


bool
isValidMp3(BEntry *entry)
{
	BNode target(entry);
	BNodeInfo info(&target);
	char buffer[B_ATTR_NAME_LENGTH];
	
	info.GetType(buffer);

	if (!strncmp(buffer, "audio/x-mpeg", 13)) 
		return true;

	return false; 
}


void
handleFile(BEntry *entry, BFile *file)
{
	// Recurse the directories
	if (recursive && entry->IsDirectory()) {
		BDirectory dir(entry);
		BEntry entryItr;
		
		dir.Rewind();
		while (dir.GetNextEntry(&entryItr, false) == B_OK) {
			BFile fileItr;
			
			handleFile(&entryItr, &fileItr);
		}
		return;
	}	

	char name[B_FILE_NAME_LENGTH];
	entry->GetName(name);
	
	// if it is not an audio/x-mpeg file, return with error
	if (!isValidMp3(entry)) {	
		fprintf(stderr, "'%s' is not an audio/x-mpeg file\n", name);
		return;
	}

	// Open File (on error, return showing error)
	if (!(file->SetTo(entry, B_READ_ONLY) == B_OK)) {
		fprintf(stderr, "could not open '%s'.\n", name);
		return;
	}

	// update file attributes

	audio_attrs audioAttrs;

	memset(&audioAttrs, 0, sizeof(audio_attrs));

	retrieveFromFile(name,file,&audioAttrs);
	writeAudioAttributes(entry,&audioAttrs);
	freeAudioAttributes(&audioAttrs);
}


void
addAttribute(BMessage *msg, char *name, char *publicName, int32 type)
{
	msg->AddString("attr:name", name); 
	msg->AddString("attr:public_name", publicName); 
	msg->AddInt32("attr:type", type); 
	msg->AddBool("attr:viewable", true); 
	msg->AddBool("attr:editable", true);
	msg->AddInt32("attr:width", (int32)50);
	msg->AddInt32("attr:alignment", B_ALIGN_LEFT); 
}


void
registerAttributes()
{
	BApplication app("application/x-vnd.pinc.id3attr");

	// First Set up the system mime type
	BMimeType mime("audio/x-mpeg");

	if (mime.InitCheck() != B_OK) {
		fputs("could not init mime type.\n",stderr);
		return;
	}
	bool installed = false;

	if (!mime.IsInstalled()) {
		mime.Install();
		installed = true;
	}
	
	// Add attributes to MIME type
	{
		BMessage msg;

		addAttribute(&msg,"Audio:Artist","Artist",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Title","Title",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Album","Album",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Track","Track",B_INT32_TYPE);
		addAttribute(&msg,"Audio:Year","Year",B_INT32_TYPE);
		addAttribute(&msg,"Audio:Comment","Comment",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Genre","Genre",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Rating","Rating",B_INT32_TYPE);
		if (frameRate)
			addAttribute(&msg,"Audio:Framerate","Framerate",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Bitrate","Bitrate",B_STRING_TYPE);
		addAttribute(&msg,"Audio:Length","Length",B_STRING_TYPE);

		mime.SetAttrInfo(&msg);
	}
	
	// Add common extensions to MIME type
	{
		char *extStrings[] = {"mp3","MP3","mp2","MP2","Mp3","Mp2"};
		const int8 numExtensions = 6;
		BMessage msg;
		char *ext;
		int i;
		
		mime.GetFileExtensions(&msg);

		for (i = 0; msg.FindString("extensions", i, (const char **)&ext) == B_OK; i++) {
			for (int j = 0; j < numExtensions; j++)
				if (extStrings[j] != NULL && !strcmp(extStrings[j],ext))
					extStrings[j] = NULL;
		}
		for (i = 0; i < numExtensions; i++) {
			if (extStrings[i] != NULL)
				msg.AddString("extensions", extStrings[i]);
		}

		mime.SetFileExtensions(&msg);
	}
	
	if (installed) {
		// the short description
		mime.SetShortDescription("mp3 Sound File");

		// set the default application	
		mime.SetPreferredApp("application/x-vnd.Be.MediaPlayer");
	}
}


void
createIndex(dev_t part,char *attr,uint32 type)
{
	if (fs_create_index(part,attr,type,0) != B_OK)
		if (errno != B_FILE_EXISTS)
			fprintf(stderr,"index creation failed for \"%s\": %s\n",attr,strerror(errno));
}


void
makeIndices()
{
	dev_t part = dev_for_path(".");

	if (part < 0) {
		fprintf(stderr, "could not get device for index creation: %s\n", strerror(part));
		return; 
	}

	createIndex(part,"Audio:Artist", B_STRING_TYPE);
	createIndex(part,"Audio:Title",  B_STRING_TYPE);
	createIndex(part,"Audio:Album",  B_STRING_TYPE);
	createIndex(part,"Audio:Track",  B_INT32_TYPE);
	createIndex(part,"Audio:Year",   B_INT32_TYPE);
	createIndex(part,"Audio:Comment",B_STRING_TYPE);
	createIndex(part,"Audio:Genre",  B_STRING_TYPE);
	createIndex(part,"Audio:Rating", B_INT32_TYPE);
	createIndex(part,"Audio:Bitrate",B_STRING_TYPE);
	createIndex(part,"Audio:Length", B_STRING_TYPE);
	if (frameRate)
		createIndex(part,"Audio:Framerate", B_STRING_TYPE);
}


void
printUsage(char *cmd)
{
	printf("Usage: %s [-srifFAST0m] <list of filenames and/or directories>\n"
			"  -r\tenter directories recursively\n"
			"  -s\tcreate \"audio/x-mpeg\" mime type and register \"Audio\" attributes\n"
			"  -i\tcreate indices for the current BFS-volume\n"
			"  -f\tattributes are overwritten if they already exist\n"
			"  -F\tuse Audio:Framerate attribute\n"
			"  -AST\tchanges every word to uppercase in A: Artist, S: Album, T: Title\n"
			"  -0\tprecede the length's minute with a zero, not a blank\n"
			"  -m\tdon't use the media kit: no bitrates, framerates and song length\n"
			, cmd);
}


int
main(int argc, char **argv)
{
	char *cmd = argv[0];
	
	if (argc == 1) {
		printUsage(cmd);
		return 1;
	}
	
	bool registerFlag = false,indexFlag = false;
	
	while (*++argv && **argv == '-') {
		for (int i = 1; (*argv)[i]; i++) {
			switch((*argv)[i]) {
				case 'r':
					recursive = true;
					break;
				case 's':
					registerFlag = true;
					break;
				case 'i':
					indexFlag = true;
					break;
				case 'f':
					force = true;
					break;
				case 'F':
					frameRate = true;
					break;
				case '0':
					leadingZero = true;
					break;
				case 'A':
					upperArtist = true;
					break;
				case 'T':
					upperTitle = true;
					break;
				case 'S':
					upperAlbum = true;
					break;
				case 'm':
					useMediaKit = false;
					break;
				default:
					printUsage(cmd);
					return(1);
			}
		}
	}
	if (registerFlag)
		registerAttributes();

	if (indexFlag)
		makeIndices();

	argv--;

	while (*++argv) {
		BEntry entry(*argv);
		BFile  file;

		if (entry.InitCheck() == B_OK)
			handleFile(&entry, &file);
		else
			fprintf(stderr, "could not find '%s'.\n", *argv); 
	}

	return 0; 
}
